/*
 * Copyright (c) 2013 - 2016, CodeWard.org
 */
#ifndef _HOSTFORMAT_H
#define _HOSTFORMAT_H

#define PORT_MAX_LEN 6

extern int hostformat_parse (const char *str, char *hostname, char *port);

#endif

